@@ -549,7 +549,7 @@ unload_foreign(File) :-
 
 
 :- if(current_predicate(win_add_dll_directory/2)).
-:- export(win_add_dll_directory/1).
+
 %!  win_add_dll_directory(+AbsDir) is det.
 %
 %   Add AbsDir to the directories where   dependent DLLs are searched on
@@ -569,6 +569,23 @@ win_add_dll_directory(Dir) :-
     getenv('PATH', Path0),
     atomic_list_concat([Path0, OSDir], ';', Path),
     setenv('PATH', Path).
+
+% Under MSYS2, the program is invoked from a bash, with the dll
+% dependencies (zlib1.dll etc.) in %MINGW_PREFIX%/bin instead of
+% the installation directory). Here we add this folder to the dll
+% search path.
+
+add_mingw_dll_directory :-
+    current_prolog_flag(msys2, true),
+    !,
+    getenv('MINGW_PREFIX', Prefix),
+    atomic_list_concat([Prefix, bin], /, Bin),
+    win_add_dll_directory(Bin).
+
+add_mingw_dll_directory.
+
+:- initialization(add_mingw_dll_directory).
+
 :- endif.
 
                  /*******************************
